/*
 * Copyright (c) 2019 Tilman Sauerbeck (tilman at code-monkey de)
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 * LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 * OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

use register;

type Reg32 = register::Register<u32>;

const PORT_BASE: u32 = 0x40049000;

pub const PORTA: u32 = PORT_BASE + 0x0000;
pub const PORTB: u32 = PORT_BASE + 0x1000;
pub const PORTC: u32 = PORT_BASE + 0x2000;
pub const PORTD: u32 = PORT_BASE + 0x3000;
pub const PORTE: u32 = PORT_BASE + 0x4000;

const SIM_SCGC5: u32 = 0x40048038;

const PORT_PCR_MUX_SHIFT: u32 = 8;
const PORT_PCR_MUX_MASK: u32 = 3 << PORT_PCR_MUX_SHIFT;

const PORT_PCR_PE: u32 = 1 << 1;
const PORT_PCR_PS: u32 = 1 << 0;

const SIM_SCGC5_PORTA: u32 = 1 << 9;
const SIM_SCGC5_PORTB: u32 = 1 << 10;
const SIM_SCGC5_PORTC: u32 = 1 << 11;
const SIM_SCGC5_PORTD: u32 = 1 << 12;
const SIM_SCGC5_PORTE: u32 = 1 << 13;

pub enum Pull {
    None,
    Up,
    Down,
}

fn pcr_offset(pin: u32) -> u32 {
    pin * 4
}

pub fn init() {
    let mut scgc5 = Reg32::new(SIM_SCGC5);

    scgc5.modify(|v| {
          v
        | SIM_SCGC5_PORTA
        | SIM_SCGC5_PORTB
        | SIM_SCGC5_PORTC
        | SIM_SCGC5_PORTD
        | SIM_SCGC5_PORTE
    });
}

pub fn set_af(port: u32, pin: u32, af: u32) {
    let mut pcr = Reg32::new(port + pcr_offset(pin));

    pcr.modify(|v| {
        (v & !PORT_PCR_MUX_MASK) | (af << PORT_PCR_MUX_SHIFT)
    });
}

pub fn set_pull(port: u32, pin: u32, pull: Pull) {
    let mut pcr = Reg32::new(port + pcr_offset(pin));

    match pull {
        Pull::None => {
            pcr.modify(|v| {
                v & !PORT_PCR_PE
            });
        },
        Pull::Up => {
            pcr.modify(|v| {
                v | PORT_PCR_PE | PORT_PCR_PS
            });
        },
        Pull::Down => {
            pcr.modify(|v| {
                (v | PORT_PCR_PE) & !PORT_PCR_PS
            });
        },
    }
}
